@@ -617,7 +617,8 @@ const Application = new Lang.Class({
                 'Cody Welsh <codyw@protonmail.com>',
                 'Isabella Ribeiro <belinhacbr@gmail.com>',
                 'Jonas Danielsson <jonas@threetimestwo.org>',
-                'Rares Visalom <rares.visalom@gmail.com>'
+                'Rares Visalom <rares.visalom@gmail.com>',
+                'Danny Mølgaard <moelgaard.dmp@gmail.com>'
             ],
             artists: [
                 'Sam Hewitt',
