@@ -37,6 +37,12 @@ const NICK_SPACING = 14; // space after nicks, matching the following elements
 
 const NICKTAG_PREFIX = 'nick';
 
+function _getColor(context) {
+    let color = context.get_color(context.get_state());
+    color.alpha *= context.get_property('opacity', context.get_state());
+    return color;
+}
+
 // Workaround for GtkTextView growing horizontally over time when
 // added to a GtkScrolledWindow with horizontal scrolling disabled
 const TextView = new Lang.Class({
@@ -59,7 +65,7 @@ const TextView = new Lang.Class({
         context.save();
         context.add_class('dim-label');
         context.set_state(Gtk.StateFlags.NORMAL);
-        this._dimColor = context.get_color(context.get_state());
+        this._dimColor = _getColor(context);
         context.restore();
 
         this.parent();
@@ -234,6 +240,8 @@ const ChatView = new Lang.Class({
     _init: function(room) {
         this.parent({ hscrollbar_policy: Gtk.PolicyType.NEVER, vexpand: true });
 
+        this.get_style_context().add_class('polari-chat-view');
+
         this._view = new TextView({ editable: false, cursor_visible: false,
                                     wrap_mode: Gtk.WrapMode.WORD_CHAR,
                                     right_margin: MARGIN });
@@ -257,6 +265,10 @@ const ChatView = new Lang.Class({
         this.connect('state-flags-changed',
                      Lang.bind(this, this._updateToplevel));
         this.connect('scroll-event', Lang.bind(this, this._onScroll));
+        this.connect('edge-reached', (w, pos) => {
+            if (pos == Gtk.PositionType.BOTTOM)
+                this._autoscroll = true;
+        });
 
         this.vadjustment.connect('value-changed',
                                  Lang.bind(this, this._onValueChanged));
@@ -284,8 +296,9 @@ const ChatView = new Lang.Class({
         this._maxNickChars = MAX_NICK_CHARS;
         this._hoveredButtonTags = [];
         this._needsIndicator = true;
-        this._pending = {};
+        this._pending = new Map();
         this._pendingLogs = [];
+        this._initialPending = [];
         this._statusCount = { left: 0, joined: 0, total: 0 };
 
         this._room.account.connect('notify::nickname', Lang.bind(this,
@@ -307,8 +320,7 @@ const ChatView = new Lang.Class({
         this._logWalker.get_events_async(NUM_INITIAL_LOG_EVENTS,
                                          Lang.bind(this, this._onLogEventsReady));
 
-        let adj = this.vadjustment;
-        this._scrollBottom = adj.upper - adj.page_size;
+        this._autoscroll = true;
 
         this._app = Gio.Application.get_default();
         PasteManager.DropTargetIface.addTargets(this, this._view);
@@ -346,7 +358,8 @@ const ChatView = new Lang.Class({
         let tagTable = buffer.get_tag_table();
         let tags = [
           { name: 'nick',
-            left_margin: MARGIN },
+            left_margin: MARGIN,
+            weight: Pango.Weight.BOLD },
           { name: 'gap',
             pixels_above_lines: 10 },
           { name: 'message',
@@ -360,10 +373,10 @@ const ChatView = new Lang.Class({
           { name: 'timestamp',
             left_margin: MARGIN,
             indent: 0,
-            weight: Pango.Weight.BOLD,
             justification: Gtk.Justification.RIGHT },
           { name: 'action',
-            left_margin: MARGIN },
+            left_margin: MARGIN,
+            style: Pango.Style.ITALIC },
           { name: 'url',
             underline: Pango.Underline.SINGLE },
           { name: 'indicator-line',
@@ -394,16 +407,16 @@ const ChatView = new Lang.Class({
         context.save();
         context.add_class('dim-label');
         context.set_state(Gtk.StateFlags.NORMAL);
-        let dimColor = context.get_color(context.get_state());
+        let dimColor = _getColor(context);
         context.restore();
 
         context.save();
         context.set_state(Gtk.StateFlags.LINK);
-        let linkColor = context.get_color(context.get_state());
-        this._activeNickColor = context.get_color(context.get_state());
+        let linkColor = _getColor(context);
+        this._activeNickColor = _getColor(context);
 
         context.set_state(Gtk.StateFlags.LINK | Gtk.StateFlags.PRELIGHT);
-        this._hoveredLinkColor = context.get_color(context.get_state());
+        this._hoveredLinkColor = _getColor(context);
         context.restore();
 
         let desaturatedNickColor = (this._activeNickColor.red +
@@ -419,7 +432,7 @@ const ChatView = new Lang.Class({
         context.save();
         context.add_class('view');
         context.set_state(Gtk.StateFlags.NORMAL);
-        this._statusHeaderHoverColor = context.get_color(context.get_state());
+        this._statusHeaderHoverColor = _getColor(context);
         context.restore();
 
         let buffer = this._view.get_buffer();
@@ -429,8 +442,6 @@ const ChatView = new Lang.Class({
             foreground_rgba: dimColor },
           { name: 'timestamp',
             foreground_rgba: dimColor },
-          { name: 'action',
-            foreground_rgba: dimColor },
           { name: 'url',
             foreground_rgba: linkColor }
         ];
@@ -460,44 +471,87 @@ const ChatView = new Lang.Class({
 
     _onLogEventsReady: function(lw, res) {
         this._hideLoadingIndicator();
+        this._fetchingBacklog = false;
 
         let [, events] = lw.get_events_finish(res);
-        this._pendingLogs = events.concat(this._pendingLogs);
+        let messages = events.map(e => this._createMessage(e));
+        this._pendingLogs = messages.concat(this._pendingLogs);
         this._insertPendingLogs();
-        this._fetchingBacklog = false;
+    },
+
+    _createMessage: function(source) {
+        if (source instanceof Tp.Message) {
+            let [text, ] = source.to_text();
+            let [id, valid] = source.get_pending_message_id();
+            return { nick: source.sender.alias,
+                     text: text,
+                     timestamp: source.get_sent_timestamp() ||
+                                source.get_received_timestamp(),
+                     messageType: source.get_message_type(),
+                     pendingId: valid ? id : undefined };
+        } else if (source instanceof Tpl.Event) {
+            return { nick: source.sender.alias,
+                     text: source.message,
+                     timestamp: source.timestamp,
+                     messageType: source.get_message_type() };
+        }
+
+        throw new Error('Cannot create message from source ' + source);
+    },
+
+    _getReadyLogs: function() {
+        if (this._logWalker.is_end())
+            return this._pendingLogs.splice(0);
+
+        let nick = this._pendingLogs[0].nick;
+        let type = this._pendingLogs[0].messageType;
+        let maxNum = this._pendingLogs.length - this._initialPending.length;
+        for (let i = 0; i < maxNum; i++)
+            if (this._pendingLogs[i].nick != nick ||
+                this._pendingLogs[i].messageType != type)
+                return this._pendingLogs.splice(i);
+        return [];
+    },
+
+    _appendInitialPending: function(logs) {
+        let pending = this._initialPending.splice(0);
+        let firstPending = pending[0];
+
+        let numLogs = logs.length;
+        let pos;
+        for (pos = numLogs - pending.length; pos < numLogs; pos++)
+            if (logs[pos].nick == firstPending.nick &&
+                logs[pos].text == firstPending.text &&
+                logs[pos].timestamp == firstPending.timestamp &&
+                logs[pos].messageType == firstPending.messageType)
+                break;
+        // Remove entries that are also in pending (if any), then
+        // add the entries from pending
+        logs.splice.apply(logs, [pos, numLogs, ...pending]);
     },
 
     _insertPendingLogs: function() {
-        if (this._pendingLogs.length == 0)
-            return;
+        let pending = this._getReadyLogs();
 
-        let index = -1;
-        let nick = this._pendingLogs[0].sender.alias;
-        let type = this._pendingLogs[0].message_type;
-        if (!this._logWalker.is_end()) {
-            for (let i = 0; i < this._pendingLogs.length; i++)
-                if (this._pendingLogs[i].sender.alias != nick ||
-                    this._pendingLogs[i].message_type != type) {
-                    index = i;
-                    break;
-                }
-        } else {
-            index = 0;
+        if (!pending.length) {
+            this._fetchBacklog();
+            return;
         }
 
-        if (index < 0)
-            return;
+        let numInitialPending = this._initialPending.length;
+        if (numInitialPending)
+            this._appendInitialPending(pending);
+
+        let indicatorIndex = pending.length - numInitialPending;
 
-        let pending = this._pendingLogs.splice(index);
         let state = { lastNick: null, lastTimestamp: 0 };
         let iter = this._view.buffer.get_start_iter();
+
         for (let i = 0; i < pending.length; i++) {
-            let message = { nick: pending[i].sender.alias,
-                            text: pending[i].message,
-                            timestamp: pending[i].timestamp,
-                            messageType: pending[i].get_message_type(),
-                            shouldHighlight: false };
-            this._insertMessage(iter, message, state);
+            this._insertMessage(iter, pending[i], state);
+
+            if (i == indicatorIndex)
+                this._setIndicatorMark(iter);
 
             if (!iter.is_end() || i < pending.length - 1)
                 this._view.buffer.insert(iter, '\n', -1);
@@ -516,10 +570,6 @@ const ChatView = new Lang.Class({
         }
     },
 
-    get _nPending() {
-        return Object.keys(this._pending).length;
-    },
-
     get max_nick_chars() {
         return this._maxNickChars;
     },
@@ -571,19 +621,17 @@ const ChatView = new Lang.Class({
     },
 
     _updateScroll: function() {
-        let adj = this.vadjustment;
-        if (adj.value == this._scrollBottom) {
-            if (this._nPending == 0) {
-                this._view.emit('move-cursor',
-                                Gtk.MovementStep.BUFFER_ENDS, 1, false);
-            } else {
-                let id = Object.keys(this._pending).sort(function(a, b) {
-                    return a - b;
-                })[0];
-                this._view.scroll_mark_onscreen(this._pending[id]);
-            }
+        if (!this._autoscroll)
+            return;
+
+        if (this._pending.size == 0) {
+            this._view.emit('move-cursor',
+                            Gtk.MovementStep.BUFFER_ENDS, 1, false);
+        } else {
+            this._autoscroll = false;
+            let mark = [...this._pending.values()].shift();
+            this._view.scroll_mark_onscreen(mark);
         }
-        this._scrollBottom = adj.upper - adj.page_size;
     },
 
     _onScroll: function(w, event) {
@@ -595,6 +643,8 @@ const ChatView = new Lang.Class({
         if (hasDeltas && dy >= 0)
             return Gdk.EVENT_PROPAGATE;
 
+        this._autoscroll = false;
+
         return this._fetchBacklog();
     },
 
@@ -621,6 +671,8 @@ const ChatView = new Lang.Class({
             keyval != Gdk.KEY_KP_Page_Up)
             return Gdk.EVENT_PROPAGATE;
 
+        this._autoscroll = false;
+
         return this._fetchBacklog();
     },
 
@@ -657,10 +709,15 @@ const ChatView = new Lang.Class({
 
     _pendingMessageRemoved: function(channel, message) {
         let [id,] = message.get_pending_message_id();
-        if (this._pending[id])
-            this._view.buffer.delete_mark(this._pending[id]);
+        let mark = this._pending.get(id);
+        if (!mark)
+            return;
+        // Re-enable auto-scrolling if this is the most recent message
+        if (this._view.buffer.get_iter_at_mark(mark).is_end())
+            this._autoscroll = true;
+        this._view.buffer.delete_mark(mark);
         this._app.withdraw_notification('pending-message-' + id);
-        delete this._pending[id];
+        this._pending.delete(id);
     },
 
     _showUrlContextMenu: function(url, button, time) {
@@ -747,6 +804,27 @@ const ChatView = new Lang.Class({
         buffer.delete(buffer.get_start_iter(), iter);
     },
 
+    _setIndicatorMark: function(iter) {
+        let lineStart = iter.copy();
+        lineStart.set_line_offset(0);
+
+        let buffer = this._view.buffer;
+        let mark = buffer.get_mark('indicator-line');
+        if (mark) {
+            let [start, end] = this._getLineIters(buffer.get_iter_at_mark(mark));
+            buffer.remove_tag(this._lookupTag('indicator-line'), start, end);
+
+            buffer.move_mark(mark, lineStart);
+        } else {
+            buffer.create_mark('indicator-line', lineStart, true);
+        }
+
+        let [start, end] = this._getLineIters(lineStart);
+        buffer.apply_tag(this._lookupTag('indicator-line'), start, end);
+
+        this._needsIndicator = false;
+    },
+
     _checkMessages: function() {
         if (!this._active || !this._toplevelFocus || !this._channel)
             return;
@@ -761,7 +839,7 @@ const ChatView = new Lang.Class({
         let buffer = this._view.get_buffer();
         for (let i = 0; i < pending.length; i++) {
             let [id,] = pending[i].get_pending_message_id();
-            let mark = this._pending[id];
+            let mark = this._pending.get(id);
             if (!mark) {
                 this._channel.ack_message_async(pending[i], null);
                 continue;
@@ -857,11 +935,8 @@ const ChatView = new Lang.Class({
             this._channelSignals.push(this._channel.connect(signal.name, signal.handler));
         }));
 
-        this._channel.dup_pending_messages().forEach(Lang.bind(this,
-            function(message) {
-                this._insertTpMessage(this._room, message);
-            }));
-        this._checkMessages();
+        let pending = this._channel.dup_pending_messages();
+        this._initialPending = pending.map(p => this._createMessage(p));
     },
 
     _onMemberRenamed: function(room, oldMember, newMember) {
@@ -913,8 +988,8 @@ const ChatView = new Lang.Class({
         this._untrackContact(member);
     },
 
-    _onMessageReceived: function(room, tpMessage) {
-        this._insertTpMessage(room, tpMessage);
+    _onMessageReceived: function(channel, tpMessage) {
+        this._insertTpMessage(tpMessage);
         this._resetStatusCompressed();
         let nick = tpMessage.sender.alias;
         let nickTag = this._lookupTag('nick' + nick);
@@ -923,8 +998,8 @@ const ChatView = new Lang.Class({
         nickTag._lastActivity = GLib.get_monotonic_time();
     },
 
-    _onMessageSent: function(room, tpMessage) {
-        this._insertTpMessage(room, tpMessage);
+    _onMessageSent: function(channel, tpMessage) {
+        this._insertTpMessage(tpMessage);
         this._resetStatusCompressed();
     },
 
@@ -1126,19 +1201,10 @@ const ChatView = new Lang.Class({
         return date.format(format);
     },
 
-    _insertTpMessage: function(room, tpMessage) {
-        let [text, flags] = tpMessage.to_text();
-
-        let message = { nick: tpMessage.sender.alias,
-                        text: text,
-                        messageType: tpMessage.get_message_type() };
-
-        let timestamp = tpMessage.get_sent_timestamp();
-        if (!timestamp)
-            timestamp = tpMessage.get_received_timestamp();
-        message.timestamp = timestamp;
-
-        message.shouldHighlight = this._room.should_highlight_message(tpMessage);
+    _insertTpMessage: function(tpMessage) {
+        let message = this._createMessage(tpMessage);
+        let shouldHighlight = this._room.should_highlight_message(message.nick,
+                                                                  message.text);
 
         this._ensureNewLine();
 
@@ -1146,9 +1212,7 @@ const ChatView = new Lang.Class({
         this._insertMessage(iter, message, this._state);
         this._trackContact(tpMessage.sender);
 
-        let [id, valid] = tpMessage.get_pending_message_id();
-
-        if (message.shouldHighlight &&
+        if (shouldHighlight &&
             !(this._toplevelFocus && this._active)) {
             let summary = '%s %s'.format(this._room.display_name, message.nick);
             let notification = new Gio.Notification();
@@ -1161,45 +1225,22 @@ const ChatView = new Lang.Class({
                                            this._room.channel_name,
                                            Utils.getTpEventTime() ]);
             notification.set_default_action_and_target('app.join-room', param);
-            this._app.send_notification('pending-message-' + id, notification);
+            this._app.send_notification('pending-message-' + message.pendingId,
+                                        notification);
         }
 
-        let buffer = this._view.get_buffer();
-        if (!valid /* outgoing */ ||
-            (this._active && this._toplevelFocus && this._nPending == 0)) {
+        if (!message.pendingId /* outgoing */ ||
+            (this._active && this._toplevelFocus && this._pending.size == 0))
             this._channel.ack_message_async(tpMessage, null);
-        } else if (message.shouldHighlight || this._needsIndicator) {
-            let iter = buffer.get_end_iter();
-
-            if (message.shouldHighlight) {
-                let mark = buffer.create_mark(null, iter, true);
-                this._pending[id] = mark;
-            }
-
-            if (this._needsIndicator) {
-                iter.set_line_offset(0);
-
-                let mark = buffer.get_mark('indicator-line');
-                if (mark) {
-                    let [start, end] = this._getLineIters(buffer.get_iter_at_mark(mark));
-                    buffer.remove_tag(this._lookupTag('indicator-line'), start, end);
-
-                    buffer.move_mark(mark, iter);
-                } else {
-                    buffer.create_mark('indicator-line', iter, true);
-                }
-
-                let [start, end] = this._getLineIters(iter);
-                buffer.apply_tag(this._lookupTag('indicator-line'), start, end);
-
-                this._needsIndicator = false;
-            }
-        }
+        else if (this._needsIndicator)
+            this._setIndicatorMark(this._view.buffer.get_end_iter());
     },
 
     _insertMessage: function(iter, message, state) {
         let isAction = message.messageType == Tp.ChannelTextMessageType.ACTION;
         let needsGap = message.nick != state.lastNick || isAction;
+        let highlight = this._room.should_highlight_message(message.nick,
+                                                            message.text);
 
         if (message.timestamp - TIMESTAMP_INTERVAL > state.lastTimestamp) {
             let tags = [this._lookupTag('timestamp')];
@@ -1242,7 +1283,7 @@ const ChatView = new Lang.Class({
             tags.push(this._lookupTag('message'));
         }
 
-        if (message.shouldHighlight)
+        if (highlight)
             tags.push(this._lookupTag('highlight'));
 
         let params = this._room.account.dup_parameters_vardict().deep_unpack();
@@ -1266,6 +1307,10 @@ const ChatView = new Lang.Class({
             pos = url.pos + name.length;
         }
         this._insertWithTags(iter, text.substr(pos), tags);
+
+        if (highlight && message.pendingId)
+            this._pending.set(message.pendingId,
+                              this._view.buffer.create_mark(null, iter, true));
     },
 
     _createUrlTag: function(url) {
