@@ -119,54 +119,34 @@ polari_create_room_id (TpAccount    *account,
 }
 
 #ifdef HAVE_STRCASESTR
+#  define FOLDFUNC(text) ((char *)(text))
 #  define MATCHFUNC(haystick,needle) strcasestr (haystick, needle)
-#  define MESSAGE_TO_TEXT(message) tp_message_to_text (message, NULL)
+#  define FREEFUNC(text)
 #else
-   static inline char *
-   message_to_casefolded_text (TpMessage *message) {
-     g_autofree char *tmp = tp_message_to_text (message, NULL);
-     return g_utf8_casefold (tmp, -1);
-   }
+#  define FOLDFUNC(text) g_utf8_casefold (text, -1)
 #  define MATCHFUNC(haystick,needle) strstr (haystick, needle)
-#  define MESSAGE_TO_TEXT(message) message_to_casefolded_text (message)
+#  define FREEFUNC(text) g_free(text)
 #endif
 
-gboolean
-polari_room_should_highlight_message (PolariRoom *room,
-                                      TpMessage *message)
+static gboolean
+match_self_nick (PolariRoom *room,
+                 const char *text)
 {
-  PolariRoomPrivate *priv;
-  TpConnection *conn;
-  TpContact *self;
-  char *text, *match;
+  PolariRoomPrivate *priv = room->priv;
+  char *folded_text, *match;
   gboolean result = FALSE;
   int len;
 
-  g_return_val_if_fail (POLARI_IS_ROOM (room), FALSE);
-
-  priv = room->priv;
-
-  if (!priv->channel)
-    return FALSE;
-  if (priv->type != TP_HANDLE_TYPE_ROOM)
-    return FALSE;
-
-  conn = tp_channel_get_connection (room->priv->channel);
-  self = tp_connection_get_self_contact (conn);
-
-  if (tp_signalled_message_get_sender (message) == self)
-    return FALSE;
-
-  text = MESSAGE_TO_TEXT (message);
+  folded_text = FOLDFUNC (text);
   len = strlen (priv->self_nick);
-  match = MATCHFUNC (text, priv->self_nick);
+  match = MATCHFUNC (folded_text, priv->self_nick);
 
   while (match != NULL)
     {
       gboolean starts_word, ends_word;
 
       /* assume ASCII nicknames, so no complex pango-style breaks */
-      starts_word = (match == text || !g_ascii_isalnum (*(match - 1)));
+      starts_word = (match == folded_text || !g_ascii_isalnum (*(match - 1)));
       ends_word = !g_ascii_isalnum (*(match + len));
 
       result = starts_word && ends_word;
@@ -175,11 +155,31 @@ polari_room_should_highlight_message (PolariRoom *room,
       match = MATCHFUNC (match + len, priv->self_nick);
     }
 
-  g_free (text);
+  FREEFUNC (folded_text);
 
   return result;
 }
 
+gboolean
+polari_room_should_highlight_message (PolariRoom *room,
+                                      const char *sender,
+                                      const char *message)
+{
+  PolariRoomPrivate *priv;
+
+  g_return_val_if_fail (POLARI_IS_ROOM (room), FALSE);
+
+  priv = room->priv;
+
+  if (priv->type != TP_HANDLE_TYPE_ROOM)
+    return FALSE;
+
+  if (match_self_nick (room, sender))
+    return FALSE;
+
+  return match_self_nick (room, message);
+}
+
 void
 polari_room_set_topic (PolariRoom *room,
                        const char *topic)
@@ -248,20 +248,27 @@ strip_color_codes (const char *string) {
 static void
 update_self_nick (PolariRoom *room)
 {
-  TpConnection *conn;
-  TpContact *self;
-
   PolariRoomPrivate *priv = room->priv;
+  const char *nick;
 
   g_clear_pointer (&priv->self_nick, g_free);
 
-  if (!room->priv->channel)
-    return;
+  if (priv->channel)
+    {
+      TpConnection *conn;
+      TpContact *self;
 
-  conn = tp_channel_get_connection (room->priv->channel);
-  self = tp_connection_get_self_contact (conn);
+      conn = tp_channel_get_connection (priv->channel);
+      self = tp_connection_get_self_contact (conn);
 
-  priv->self_nick = polari_util_get_basenick (tp_contact_get_alias (self));
+      nick = tp_contact_get_alias (self);
+    }
+  else
+    {
+      nick = tp_account_get_nickname (priv->account);
+    }
+
+  priv->self_nick = polari_util_get_basenick (nick);
 }
 
 static void
@@ -476,6 +483,8 @@ polari_room_set_account (PolariRoom *room,
 
   if (g_set_object (&priv->account, account))
     g_object_notify_by_pspec (G_OBJECT (room), props[PROP_ACCOUNT]);
+
+  update_self_nick (room);
 }
 
 static void
